module Aspect
  # Transforms a `Hash` into human language.
  #
  # @example Getting using `Symbol`.
  #   require "aspect/message_transform"
  #
  #   message = Aspect::MessageTransform.new(login: "Click to login", body: "Lorem ipsum")
  #
  #   p message.to_a(:login, :body) # => ["Click to login", "Lorem ipsum"]
  #   p message.to_h(:login, :body) # => { login: "Click to login", body: "Lorem ipsum" }
  #   p message.to_s(:login, :body) # => "Click to login, Lorem ipsum"
  #
  #   p message.to_a(:body) # => ["Lorem ipsum"]
  #   p message.to_h(:body) # => { body: "Lorem ipsum" }
  #   p message.to_s(:body) # => "Lorem ipsum"
  # @example Getting using `Hash`.
  #   require "aspect/message_transform"
  #
  #   message = Aspect::MessageTransform.new(login: "Click to login", body: "Lorem ipsum")
  #
  #   p message.to_a(login: true, body: true) # => ["Click to login", "Lorem ipsum"]
  #   p message.to_h(login: true, body: true) # => { login: "Click to login", body: "Lorem ipsum" }
  #   p message.to_s(login: true, body: true) # => "Click to login, Lorem ipsum"
  #
  #   p message.to_a(login: false, body: true) # => ["Lorem ipsum"]
  #   p message.to_h(login: false, body: true) # => { body: "Lorem ipsum" }
  #   p message.to_s(login: false, body: true) # => "Lorem ipsum"
  #
  #   p message.to_a(body: true) # => ["Lorem ipsum"]
  #   p message.to_h(body: true) # => { body: "Lorem ipsum" }
  #   p message.to_s(body: true) # => "Lorem ipsum"
  # @example Output formatting and separator.
  #   require "aspect/message_transform"
  #
  #   message = Aspect::MessageTransform.new(login: "Click to login", body: "Lorem ipsum")
  #
  #   message.separator = " - "
  #   message.formatter { |value| "<div>#{value}</div>" }
  #   # Or (can be anything responding to #call(value)):
  #   # message.formatter = -> { |value| "<div>#{value}</div>" }
  #
  #   p message.to_a(:login, :body) # => ["<div>Click to login</div>", "<div>Lorem ipsum</div>"]
  #   p message.to_h(:login, :body) # => { login: "<div>Click to login</div>", body: "<div>Lorem ipsum</div>" }
  #   p message.to_s(:login, :body) # => "<div>Click to login</div> - <div>Lorem ipsum</div>"
  # @example Message interpolation.
  #   require "aspect/message_transform"
  #
  #   message = Aspect::MessageTransform.new(hello: "Hello %{hello}!")
  #
  #   p message.to_s(hello: "world") # => "Hello world!"
  # @example Getting with a multi-level `Hash`.
  #   require "aspect/message_transform"
  #
  #   message = Aspect::MessageTransform.new(header: "My App", body: { title: "Some Page", hello: "Hello %{hello}!" })
  #
  #   p message.to_s(header: true)             # => "My App"
  #   p message.to_s(body: { title: true })    # => "Some Page"
  #   p message.to_s(body: { hello: "world" }) # => "Hello world!"
  # @example Getting with `String` notation.
  #   require "aspect/message_transform"
  #
  #   message = Aspect::MessageTransform.new(header: "My App", body: { title: "Some Page", hello: "Hello %{hello}!" })
  #
  #   p message.to_s("header")                              # => "My App"
  #   p message.to_s("body.title")                          # => "Some Page"
  #   p message.to_s("body.hello" => "world")               # => "Hello world!"
  #   p message.to_a("body.title", "body.hello" => "world") # => ["Some Page", "Hello world!"]
  # @example Internationalization (I18N) using multi-level `Hash`, compatible with {Verifier} output.
  #   require "aspect/message_transform"
  #
  #   dictionary = {
  #     en: {
  #       presence: "%{attribute} must not be blank",
  #       greater_than_or_equal_to: "%{attribute} must be greater than or equal to %{greater_than_or_equal_to}"
  #     },
  #     es: {
  #       presence: "%{attribute} no debe estar en blanco",
  #       greater_than_or_equal_to: "%{attribute} debe ser mayor que o igual a %{greater_than_or_equal_to}"
  #     }
  #   }
  #
  #   error_message = Aspect::MessageTransform.new(dictionary)
  #
  #   error_message.to_s(en: { attribute: "age", presence: true, greater_than_or_equal_to: 18 }) # => "age must not be blank, age must be greater than or equal to 18"
  #   error_message.to_s(es: { attribute: "edad", presence: true, greater_than_or_equal_to: 18 }) # => "edad no debe estar en blanco, edad debe ser mayor que o igual a 18"
  class MessageTransform
    def initialize(expectations={})
      @expectations = expectations.to_h
    end

    def inspect
      memory_location = (object_id << 1).to_s(16)

      "#<#{self.class}>:0x#{memory_location} #{@expectations.inspect}>"
    end
  end
end
