if Kernel.respond_to?(:require)
  require "aspect/has_attributes"
  require "aspect/verifier/check"
end

module Aspect
  # Verify a set of checks on an object.
  #
  # The result is a `Hash` where the keys are the check name and the value is the expectation or `nil` when
  # the object is verified to be valid.
  #
  # This allows for it to easily be transformed into messages using {Aspect::MessageTransform}, if needed.
  #
  # @example Simple boolean checks.
  #   require "aspect/verifier"
  #
  #   verifier = Aspect::Verifier.new(:presence, :integer)
  #
  #   verifier.verify(nil)     # => { presence: true, integer: true } # Invalid
  #   verifier.verify("")      # => { presence: true, integer: true } # Invalid
  #   verifier.verify("hello") # => { integer: true }                 # Invalid
  #   verifier.verify("24")    # => nil                               # Valid
  # @example Simple checks with arguments.
  #   require "aspect/verifier"
  #
  #   verifier = Aspect::Verifier.new(integer: true, greater_than: 10)
  #
  #   verifier.verify(10) # => { greater_than: 10 }                     # Invalid
  #   verifier.verify("hello") # => { integer: true, greater_than: 10 } # Invalid
  #   verifier.verify("24") # => nil                                    # Valid
  # @example Break on fail.
  #   require "aspect/verifier"
  #
  #   verifier = Aspect::Verifier.new(integer: true, greater_than: 10)
  #
  #   verifier.verify("hello") # => { integer: true, greater_than: 10 }
  #   verifier.verify("hello", break_on_fail: true) # => { integer: true }
  # @example Fake application usage example
  #   require "aspect/verifier"
  #
  #   age_verifier = Aspect::Verifier.new(:integer, greater_than_or_equal_to: 18)
  #
  #   loop do
  #     print "age? "
  #
  #     input = gets
  #     error = age_verifier.verify(input, break_on_fail: true)
  #
  #     if error
  #       puts "  Must give a number." if error[:integer]
  #       puts "  Must be #{error[:greater_than_or_equal_to]} or older." if error[:greater_than_or_equal_to]
  #     else
  #       puts "  Welcome, appropriately-aged user!"
  #
  #       break
  #     end
  #   end
  #
  #   # Command line:
  #   #   age? foobar
  #   #     Must give a number.
  #   #   age? 17
  #   #     Must be 18 or older.
  #   #   age? 25
  #   #     Welcome, appropriately-aged user!
  # @example Fake application usage example with {Aspect::MessageTransform}
  #   require "aspect/verifier"
  #   require "aspect/message_transform"
  #
  #   age_verifier = Aspect::Verifier.new(:integer, greater_than_or_equal_to: 18)
  #   message = Aspect::MessageTransform.new(
  #     welcome: "Welcome, appropriately-aged user!",
  #     error: {
  #       integer: "Must give a number.",
  #       greater_than_or_equal_to: "Must be %{greater_than_or_equal_to} or older."
  #     }
  #   )
  #
  #   loop do
  #     print "age? "
  #
  #     input = gets
  #     error = age_verifier.verify(input, break_on_fail: true)
  #
  #     print "  "
  #     if error
  #       puts message.to_s(error: error)
  #     else
  #       puts message.to_s(:welcome)
  #
  #       break
  #     end
  #   end
  #
  #   # Command line:
  #   #   age? foobar
  #   #     Must give a number.
  #   #   age? 17
  #   #     Must be 18 or older.
  #   #   age? 25
  #   #     Welcome, appropriately-aged user!
  class Verifier
    include HasAttributes

    # @example Multiple flags, for boolean checks.
    #   Aspect::Verifier.new(:presence, :integer)
    # @example `Array` of flags, for boolean checks.
    #   Aspect::Verifier.new([:presence, :integer])
    # @example Passing a `Hash`, for non-boolean checks.
    #   Aspect::Verifier.new(presence: true, integer: true, greater_than: 10)
    # @example Mixing both boolean and non-boolean checks.
    #   Aspect::Verifier.new(:presence, :integer, greater_than: 10)
    def initialize(*arguments)
      attributes = {}

      if arguments.length == 1 && arguments[0].respond_to?(:to_h)
        attributes = arguments[0]
      else
        arguments = arguments[0] if arguments.length == 1 && arguments[0].is_a?(Array)

        arguments.collect(&:to_sym).each { |check| attributes[check] = true }
      end

      update_attributes(attributes)
    end

    # Validate a value.
    #
    # @param value
    # @param [#to_h] options
    # @option options [Boolean] :break_on_fail Set to `true` to stop running checks as soon as one fails.
    # @return [<Symbol, Object>] The result is a `Hash` where the keys are the check name and the value is the expectation or `nil` when the object is verified to be valid.
    def validate(value, options={})
      options = { break_on_fail: false }.merge(options.to_h)
      errors = {}

      Check.registry.each do |check_name, check|
        check_iv = instance_variable_get("@#{check_name}")

        errors[check_name] = check_iv if check_iv && !check.call(value, check_iv)

        break if options[:break_on_fail]
      end

      errors
    end

    # @method presence=
    # Set whether to check for presence.
    #
    # Objects that are `nil`, or respond to `empty?` and return `true`, are considered blank.
    #
    # @param [Boolean] value
    # @return [Boolean]

    # @method presence?
    # Get whether to check for presence.
    #
    # @return [Boolean]
    attribute(:presence, query: true)

    # @method integer=
    # Set whether to check for `Integer`.
    #
    # The values `123` and `"123"` will both pass, but `123.4` and `"123.4"` will both fail.
    #
    # @param [Boolean] value
    # @return [Boolean]

    # @method integer?
    # Get whether to check for `Integer`.
    #
    # @return [Boolean]
    attribute(:integer, query: true)

    # @method float=
    # Set whether to check for `Float`.
    #
    # The values `123.4` and `"123.4"` will both pass.
    #
    # @param [Boolean] value
    # @return [Boolean]

    # @method float?
    # Get whether to check for `Float`.
    #
    # @return [Boolean]
    attribute(:float, query: true)

    # @method equal_to=
    # Set the object to check equality against.
    #
    # @param [nil, Object] value
    # @return [nil, Object]

    # @method equal_to
    # Get whether to for equality against a given object.
    #
    # @return [Boolean]
    attribute(:equal_to)

    attribute(:greater_than) { |value| value.to_f }

    attribute(:greater_than_or_equal_to) { |value| value.to_f }

    attribute(:less_than) { |value| value.to_f }

    attribute(:less_than_or_equal_to) { |value| value.to_f }

    # TODO: documentation for equal_to, greater_than, greater_than_or_equal_to, less_than, less_than_or_equal_to
  end
end
