require "aspect/has_registry" if Kernel.respond_to?(:require)

module Aspect
  class Verifier
    # A single verification check.
    #
    # To define a new check, you must first register it, then define an attribute for it on the {Aspect::Verifier} class.
    #
    # If the check's block returns a falsey value or raises an error, the check is considered failed.
    # When a truthy value is return, the check has passed.
    #
    # @example Basic usage.
    #   require "aspect/verifier/check"
    #
    #   check = Aspect::Verifier::Check[:greater_than_or_equal_to]
    #
    #   loop do
    #     age = gets
    #
    #     if check.call(age, 18)
    #       puts "Welcome!"
    #
    #       break
    #     else
    #       puts "Must be 18 or older"
    #     end
    #   end
    # @example Defining a new {Verifier} check.
    #   require "aspect/verifier"
    #
    #   Aspect::Verifier::Check.register(:length_of_at_least) { |value, expectation| value.length >= expectation }
    #   Aspect::Verifier.attribute(:length_of_at_least) { |value| value.to_i }
    #
    #   verifier = Aspect::Verifier.new(length_of_at_least: 3)
    #
    #   verifier.verify("12") # => { length_of_at_least: 3 } # Invalid
    #   verifier.verify("123") # => nil                      # Valid
    class Check
      extend HasRegistry

      class << self
        # Get a registered `Check` instance by it's name.
        #
        # @param [#to_sym] name
        # @return [Check]
        def [](name)
          registry[name.to_sym]
        end
      end

      def initialize(&block)
        self.block = block
      end

      # Get or set the block for the check.
      #
      # @return [Proc]
      def block(&block)
        self.block = block if block_given?

        @block
      end

      # Set the block for the check.
      #
      # @param [Proc] value
      # @return [Proc]
      def block=(value)
        if block_given?
          raise ArgumentError, "block must have 1 block argument" unless value.arity == 1

          @block = value
        end

        @block
      end

      # Call the block.
      #
      # @param [Object] value
      # @return [Boolean]
      def call(value, expectation)
        !!@block.call(value, expectation) rescue false
      end

      register(:equal_to) do |value, expectation|
        value == expectation
      end

      register(:presence) do |value|
        value = value.is_a?(String) ? value.gsub(/\s+/, "") : value

        value.nil? || value.respond_to?(:empty?) && value.empty?
      end

      register(:integer) { |value| Kernel.Integer(value.to_s) }

      register(:float) { |value| Kernel.Float(value.to_s) }

      register(:greater_than) do |value, expectation|
        value = Kernel.Float(value.to_s)

        value > expectation
      end

      register(:greater_than_or_equal_to) do |value, expectation|
        value = Kernel.Float(value.to_s)

        value >= expectation
      end

      register(:less_than) do |value, expectation|
        value = Kernel.Float(value.to_s)

        value < expectation
      end

      register(:less_than_or_equal_to) do |value, expectation|
        value = Kernel.Float(value.to_s)

        value <= expectation
      end
    end
  end
end
